/*
    Yin 3D Game Engine
    Copyright (C) 2020 Mark Sowden <hogsy@oldtimes-software.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
/* Copyright (C) 2020 Mark Sowden <markelswo@gmail.com>
 * Project Yin
 * */

#include "yin.h"
#include "actor.h"
#include "renderer/renderer.h"

#define PLAYER_VIEW_OFFSET      75.0f
#define PLAYER_CROUCH_OFFSET    45.0f

#define PLAYER_TURN_SPEED    2.0f
#define PLAYER_WALK_SPEED    2.0f
#define PLAYER_RUN_SPEED     4.0f
#define PLAYER_MAX_VELOCITY  PLAYER_RUN_SPEED
#define PLAYER_MIN_VELOCITY  0.5f

#define PLAYER_MAX_PITCH    85.0f
#define PLAYER_MIN_PITCH    -85.0f

static const PLVector3 playerBoundsMins = PLVector3( -16.0f, 0.0f, -16.0f );
static const PLVector3 playerBoundsMaxs = PLVector3( 16.0f, 90.0f, 16.0f );

typedef struct APlayer {
	PLVector3 ulViewPos;
	PLVector3 urViewPos;
	PLVector3 llViewPos;
	PLVector3 lrViewPos;

	PLVector3 centerView; /* center */

	float forwardVelocity;
	float viewBob;

	GfxCamera *eyeCamera;
} APlayer;

GfxCamera *Player_GetCamera( Actor *self ) {
	APlayer *playerData = Act_GetUserData( self );
	if ( playerData == NULL ) {
		return NULL;
	}

	return playerData->eyeCamera;
}

static void Player_CalculateViewFrustum( Actor *self ) {
	APlayer *playerData = Act_GetUserData( self );

	PLVector3 forward, left;
	plAnglesAxes( PLVector3( 0, Act_GetAngle( self ), 0 ), &left, NULL, &forward );

	PLVector3 curPos = Act_GetPosition( self );
	curPos.y += Act_GetViewOffset( self );

	playerData->centerView = plAddVector3( curPos, plScaleVector3f( forward, 1000.0f ) );

	playerData->llViewPos = plAddVector3( curPos, plScaleVector3f( left, 64.0f ) );
	playerData->lrViewPos = plSubtractVector3( curPos, plScaleVector3f( left, 64.0f ) );

	/* in future, set this up properly relative to view */
}

/**
 * Ensure the 2D point provided is forward of the player's position
 */
bool Player_IsPointVisible( Actor *self, const PLVector2 *point ) {
	if ( Act_GetType( self ) != ACTOR_PLAYER ) {
		return false;
	}

	APlayer *playerData = Act_GetUserData( self );
	if ( playerData == NULL ) {
		return false;
	}

	PLVector2 lineStart = PLVector2( playerData->llViewPos.x, playerData->llViewPos.z );
	PLVector2 lineEnd = PLVector2( playerData->lrViewPos.x, playerData->lrViewPos.z );

	/* in future, set this up properly relative to view */

	float d = plTestPointLinePosition( point, &lineStart, &lineEnd );
	if( d > 0.0f ) {
		return false;
	}

	return true;
}

/* move this somewhere else... */
static unsigned int numPlayers = 0;

void Player_Spawn( Actor *self ) {
	APlayer* playerData = Sys_calloc( 1, sizeof( APlayer ) );
	Act_SetUserData( self, playerData );

	if ( numPlayers == 0 ) { /* local player */
		playerData->eyeCamera = Gfx_CreateCamera( VIEW_PERSPECTIVE_EYE, Act_GetPosition( self ), PLVector3( 0, Act_GetAngle( self ), 0 ), Engine_GetMainWindow() );
		playerData->eyeCamera->parentActor = self;
	}

	Act_SetBounds( self, playerBoundsMins, playerBoundsMaxs );

	Act_SetViewOffset( self, PLAYER_VIEW_OFFSET );
	Player_CalculateViewFrustum( self );

	numPlayers++;
}

void Player_Tick( Actor *self, void *userData ) {
	PLVector3 curOrigin = Act_GetPosition( self );
	PLVector3 curVelocity = Act_GetVelocity( self );

	float nAngle = Act_GetAngle( self );
	if ( g_system.GetButtonState( INPUT_LEFT ) ) {
		nAngle += PLAYER_TURN_SPEED;
	} else if ( g_system.GetButtonState( INPUT_RIGHT ) ) {
		nAngle -= PLAYER_TURN_SPEED;
	}
	Act_SetAngle( self, nAngle );

	if ( g_system.GetButtonState( INPUT_A ) ) {
		curVelocity.y += 10.0f;
	}

	static const float incAmount = 0.25f;
	APlayer *playerData = ( APlayer * ) userData;
	if ( g_system.GetButtonState( INPUT_UP ) ) {
		playerData->forwardVelocity += incAmount;
	} else if ( g_system.GetButtonState( INPUT_DOWN ) ) {
		playerData->forwardVelocity -= incAmount;
	} else if ( playerData->forwardVelocity != 0.0f ) {
		playerData->forwardVelocity = playerData->forwardVelocity > 0 ? playerData->forwardVelocity - incAmount : playerData->forwardVelocity + incAmount;
		if( playerData->forwardVelocity < 0.1f && playerData->forwardVelocity > -0.1f ) {
			playerData->forwardVelocity = 0.0f;
		}
	}

	float viewPitch = Act_GetViewPitch( self );
	if ( g_system.GetKeyState( 'q' ) ) {
		viewPitch += 1.0f;
	} else if ( g_system.GetKeyState( 'e' ) ) {
		viewPitch -= 1.0f;
	}

	/* clamp the view pitch */
	if ( viewPitch < PLAYER_MIN_PITCH ) {
		viewPitch = PLAYER_MIN_PITCH;
	} else if ( viewPitch > PLAYER_MAX_PITCH ) {
		viewPitch = PLAYER_MAX_PITCH;
	}

	Act_SetViewPitch( self, viewPitch );

	/* clamp the velocity as necessary */
	float maxVelocity = g_system.GetButtonState( INPUT_LEFT_STICK ) ? PLAYER_RUN_SPEED : PLAYER_WALK_SPEED;
	playerData->forwardVelocity = plClamp( -maxVelocity, playerData->forwardVelocity, maxVelocity );

	curVelocity = plAddVector3( curVelocity, plScaleVector3f( Act_GetForward( self ), playerData->forwardVelocity ) );
	Act_SetVelocity( self, &curVelocity );

	Player_CalculateViewFrustum( self );

	/* apply view bob */
	float velocityVector = plVector3Length( curVelocity );
	playerData->viewBob += ( sinf( Engine_GetNumTicks() / 5.0f ) / 10.0f ) * velocityVector;

	float viewOffset = curOrigin.y + PLAYER_VIEW_OFFSET;
	if ( g_system.GetKeyState( 'c' ) ) {
		viewOffset = curOrigin.y + PLAYER_CROUCH_OFFSET;
	}

	Act_SetViewOffset( self, viewOffset /*+ playerData->viewBob*/ );
	Act_SetPosition( self, &curOrigin );
}

void Player_Collide( Actor *self, Actor *other, void *userData ) {
	Monster_Collide( self, other, userData );

	APlayer *playerData = (APlayer *)userData;
	playerData->forwardVelocity = ( playerData->forwardVelocity / 2.0f ) * -1.0f;
}
