/*
    Yin 3D Game Engine
    Copyright (C) 2020 Mark Sowden <hogsy@oldtimes-software.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
/* Copyright (C) 2020 Mark Sowden <markelswo@gmail.com> */

#include <PL/platform.h>
#include <PL/platform_filesystem.h>
#include <PL/platform_image.h>

#include "pkgman.h"
#include "parser.h"

#define PKG_USE_COMPRESSION
#if defined( PKG_USE_COMPRESSION )
#include "miniz.h"
#endif

/* PkgMan, the shitty package generator! */

#define PKG_IDENTIFIER "PKG2"

typedef struct PkgHeader {
	char 		identifier[ 4 ];
	uint32_t	numFiles;
} PkgHeader;
static PkgHeader packageHeader = {
		.identifier = PKG_IDENTIFIER,
		.numFiles	= 0
};

static FILE *fileOutPtr = NULL;
static char outputPath[ 32 ] = { '\0' };

static void Pkg_AddFile( const char *filePath ) {
	Print( "Adding %s...\n", filePath );

	PLFile *filePtr = plOpenFile( filePath, true );
	if ( filePtr == NULL ) {
		Error( "Failed to add file \"%s\"!\nPL: %s\n", filePath, plGetError() );
	}

	const uint8_t *data = plGetFileData( filePtr );
	unsigned long fileLength = plGetFileSize( filePtr );

#if defined( PKG_USE_COMPRESSION )
	/* now compress it */
	unsigned long compressedLength = mz_compressBound( fileLength );
	uint8_t *compressedData = malloc( compressedLength );
	int status = mz_compress( compressedData, &compressedLength, data, fileLength );
	if ( status != Z_OK ) {
		Error( "Failed to compress the given file, \"%s\"!\n", filePath );
	}

	/* check if it's actually worth it... */
	if ( compressedLength > fileLength ) {
		compressedLength = fileLength;
	}
#else
	unsigned long compressedLength = fileLength;
	const uint8_t *compressedData = data;
#endif

	/* write the index header */
	uint8_t nameLength = ( uint8_t ) strlen( filePath );
	fwrite( &nameLength, sizeof( uint8_t ), 1, fileOutPtr );
	fwrite( filePath, sizeof( char ), nameLength, fileOutPtr );
	fwrite( &fileLength, sizeof( unsigned long ), 1, fileOutPtr );
	fwrite( &compressedLength, sizeof( unsigned long ), 1, fileOutPtr );

	/* and now write out the file itself */
	if ( compressedLength == fileLength ) {
		fwrite( data, 1, fileLength, fileOutPtr );
	} else {
		fwrite( compressedData, 1, compressedLength, fileOutPtr );
	}

	free( compressedData );

	plCloseFile( filePtr );

	packageHeader.numFiles++;
}

/**
 * Callback used by ScanDirectory function.
 */
static void Pkg_AddFileCallback( const char *filePath, void *userData ) {
	Pkg_AddFile( filePath );
}

static void ParseScript( const char *buffer, size_t length ) {
	const char *curPos = buffer;
	while( curPos != NULL && *curPos != '\0' ) {
		if ( *curPos == ';' ) { /* comment */
			curPos = EZP_SkipLine( curPos );
			continue;
		} else if ( strncmp( curPos, "output ", 7 ) == 0 ) { /* set output dir */
			curPos += 7;
			curPos = EZP_SkipSpaces( curPos );

			if ( outputPath[ 0 ] != '\0' ) {
				Error( "Output was already specified previously in script!\n" );
			}

			/* fetch the output path we want */
			curPos = EZP_ReadString( curPos, outputPath, sizeof( outputPath ) );
			if ( curPos == NULL ) {
				Error( "Output path did not fit into destination!\n" );
			}

			Print( "OUTPUT: %s\n", outputPath );

			fileOutPtr = fopen( outputPath, "wb" );
			if ( fileOutPtr == NULL ) {
				Error( "Failed to open \"%s\" for writing!\n", outputPath );
			}

			/* write out the file header */
			fwrite( &packageHeader, sizeof( PkgHeader ), 1, fileOutPtr );

			continue;
		} else if ( strncmp( curPos, "add ", 4 ) == 0 ) { /* add file */
			curPos += 4;
			curPos = EZP_SkipSpaces(curPos);

			char filePath[PL_SYSTEM_MAX_PATH];
			curPos = EZP_ReadString(curPos, filePath, sizeof(filePath));
			if ( curPos == NULL ) {
				Error( "File path did not fit into destination!\n" );
			}

			Pkg_AddFile(filePath);
			continue;
		} else if ( strncmp( curPos, "dir ", 4 ) == 0 ) {
			curPos += 4;
			curPos = EZP_SkipSpaces( curPos );

			char directory[ PL_SYSTEM_MAX_PATH ];
			curPos = EZP_ReadString( curPos, directory, sizeof( directory ) );
			if ( curPos == NULL ) {
				Error( "Directory path did not fit into destination!\n" );
			}

			char extension[ 8 ];
			curPos = EZP_ReadString( curPos, extension, sizeof( extension ) );
			if ( curPos == NULL ) {
				Error( "Extension did not fit into destination!\n" );
			}

			plScanDirectory( directory, extension, Pkg_AddFileCallback, false, NULL );
			continue;
		}

		curPos++;
	}

	fseek( fileOutPtr, SEEK_SET, 0 );
	fwrite( &packageHeader, sizeof( PkgHeader ), 1, fileOutPtr );
	fclose( fileOutPtr );
}

int main( int argc, char **argv ) {
	plInitialize( argc, argv );

	plRegisterStandardImageLoaders( PL_IMAGE_FILEFORMAT_ALL );

	Print( "Package Manager\nCopyright (C) 2020 Mark E Sowden <markelswo@gmail.com>\n" );
	if ( argc < 2 ) {
		Print( "Please provide a package script!\nExample: pkgman myscript.txt\n" );
		return EXIT_SUCCESS;
	}

	/* open the file and read it all into memory */
	const char *input = argv[ 1 ];
	PLFile *filePtr = plOpenFile( input, true );
	if ( filePtr == NULL ) {
		Error( "Failed to open \"%s\"!\nPL: %s\n", argv[ 1 ], plGetError() );
	}

	/* now fetch the buffer and length, and throw it to our parser */
	const char *buffer = ( const char* ) plGetFileData( filePtr );
	size_t length = plGetFileSize( filePtr );
	ParseScript( buffer, length );

	plCloseFile( filePtr );

	Print( "Done!\n" );
}
